# Copyright (c) Kuba Szczodrzyński 2022-05-11.

from .base import LabelShape, Shape
from .circle import Circle
from .fill_style import FillStyle
from .group import ShapeGroup
from .rect import Rect
from .text import Text

__all__ = [
    "Shape",
    "Circle",
    "FillStyle",
    "ShapeGroup",
    "Rect",
    "Text",
    "LabelShape",
]



#  Copyright (c) Kuba Szczodrzyński 2023-6-3.

import json

from ltchiptool.util.dict import get

MARKER = "\ufffc"
LF = "\n"


def find_marker(obj: dict | list) -> str | None:
    if isinstance(obj, dict):
        for key, value in obj.items():
            if MARKER in key:
                return key.replace(MARKER, "")
            if isinstance(value, (dict, list)):
                ret = find_marker(value)
                if ret:
                    return f"{key}.{ret}"
            if isinstance(value, str) and MARKER in value:
                return key
    elif isinstance(obj, list):
        for key, value in enumerate(obj):
            if isinstance(value, (dict, list)):
                ret = find_marker(value)
                if ret:
                    return f"{key}.{ret}"
            if isinstance(value, str) and MARKER in value:
                return str(key)
    return None


def jsonpath(s: str, p: int) -> str | None:
    before = s[0:p]
    after = s[p:]
    if (
        before.endswith("\\")
        and after.startswith("\\")
        and not before.endswith("\\\\")
        and not after.startswith("\\\\")
    ):
        p -= 1
        before = s[0:p]
        after = s[p:]
    try:
        # try deserializing with added marker
        s = before + MARKER + after
        # print(s)
        obj = json.loads(s)
        return find_marker(obj)
    except json.JSONDecodeError:
        pass

    before, _, line_start = before.rpartition(LF)
    line_end, _, after = after.partition(LF)
    try:
        dist_start = len(line_start) - line_start.rindex('"')
    except ValueError:
        dist_start = len(s)
    try:
        dist_end = line_end.index('"') + 1
    except ValueError:
        dist_end = len(s)
    if dist_start <= dist_end:
        s = (
            before
            + LF
            + line_start[:-dist_start]
            + MARKER
            + line_start[-dist_start:]
            + line_end
            + LF
            + after
        )
    elif dist_end < dist_start:
        s = (
            before
            + LF
            + line_start
            + line_end[:dist_end]
            + MARKER
            + line_end[dist_end:]
            + LF
            + after
        )
    try:
        # print(s)
        obj = json.loads(s)
        return find_marker(obj)
    except json.JSONDecodeError:
        pass

    before = before + LF + line_start
    after = line_end + LF + after
    try:
        dist_before = len(before) - before.rindex('"')
    except ValueError:
        dist_before = len(s)
    try:
        dist_after = after.index('"') + 1
    except ValueError:
        dist_after = len(s)
    if dist_before <= dist_after:
        s = before[:-dist_before] + MARKER + before[-dist_before:] + after
    elif dist_after < dist_before:
        s = before + after[:dist_after] + MARKER + after[dist_after:]
    try:
        # print(s)
        obj = json.loads(s)
        return find_marker(obj)
    except json.JSONDecodeError:
        pass
    return None


def test_jsonpath():
    json_data = """
    {
        "mask_black": {
            "fill": {
                "lgrad": ["1,0", "#4D4D4D", "0,1", "#0F0F0F"]
            },
            "stroke": {
                "color": "#B5A739",
                "width": 0.1
            }
        }
    }
    """

    obj = json.loads(json_data)
    for i in range(0, len(json_data)):
        path = jsonpath(json_data, i)
        value = get(obj, path)
        print(f"{i=}", obj, path, value)


if __name__ == "__main__":
    test_jsonpath()



# Copyright (c) Kuba Szczodrzyński 2022-05-11.

import json
from copy import deepcopy
from typing import Any

from pydantic.color import Color
from svgwrite import Drawing

from ..mixins import HasId, HasVars
from ..models.enums import LabelDir, RoleType, ShapeType
from ..utils import EvalFloat, Model, splitxy, var
from ..vector import V


def remap(shape: dict):
    tuples = ["pos", "size"]
    for tpl in tuples:
        if tpl in shape:
            shape[tpl] = splitxy(shape[tpl])
    if "fill" in shape:
        if "lgrad" in shape["fill"]:
            shape["fill"]["lgrad"][0] = splitxy(shape["fill"]["lgrad"][0])
            shape["fill"]["lgrad"][2] = splitxy(shape["fill"]["lgrad"][2])
    return shape


class Shape(Model, HasId):
    base_id: str = None
    pos: V

    # for pad labels
    label_dir: LabelDir = None
    label_size: EvalFloat = None

    def draw(self, dwg: Drawing):
        raise NotImplementedError()

    @staticmethod
    def deserialize(
        core,
        parent: HasId | HasVars | Any,
        data: dict,
        # offset: tuple[float, float] = None,
    ) -> "Shape":
        # do not modify source object
        data = deepcopy(data)
        # allow includes without specified type
        if "type" not in data and "name" in data:
            data["type"] = "include"

        # prepend id with parent id path
        if isinstance(parent, HasId):
            if parent.fullid and "id" in data:
                data["base_id"] = data["id"]
                data["id"] = parent.fullid + "." + data["id"]

        # merge parent and child vars
        vars = {}
        if isinstance(parent, HasVars):
            vars = dict(parent.vars)
        if "vars" in data:
            vars |= data["vars"]

        if vars:
            # ugly way to replace all vars in input JSON
            data = json.dumps(data)
            data = var(data, vars)
            data = json.loads(data)
            # build presets with current object's vars
            presets = core.build_presets(vars)
        else:
            # presets without vars
            presets = core.presets

        # apply shape preset(s)
        if "preset" in data:
            data |= presets[data["preset"]]
        if "presets" in data:
            for preset in data["presets"]:
                data |= presets[preset]

        # remap strings to tuples, etc.
        data = remap(data)

        shape_type = ShapeType(data["type"])
        ctor = core.shape_ctors[shape_type]
        ctor.update_forward_refs()
        return ctor(
            **data,
            core=core,
            parent=parent,
        )

    def move(self, vec: V):
        self.pos += vec

    @property
    def anchor(self) -> V:
        return self.pos

    @property
    def pos1(self) -> V:
        return V(self.x1, self.y1)

    @property
    def pos2(self) -> V:
        return V(self.x2, self.y2)

    @property
    def size(self) -> V:
        return V(self.width, self.height)

    @property
    def center(self) -> V:
        return self.pos1 + self.size / 2

    @property
    def x1(self) -> float:
        return self.pos.x

    @property
    def y1(self) -> float:
        return self.pos.y

    @property
    def x2(self) -> float:
        raise NotImplementedError()

    @property
    def y2(self) -> float:
        raise NotImplementedError()

    @property
    def width(self) -> float:
        return self.x2 - self.x1

    @property
    def height(self) -> float:
        return self.y2 - self.y1


class LabelShape(Shape):
    role_type: RoleType
    padding: V = V(0.1, 0.2)
    ratio: float
    color: Color

    @property
    def dirv(self):
        return -1 if self.label_dir == LabelDir.LEFT else 1

    @property
    def width(self) -> float:
        return (self.label_size * self.ratio) - self.padding.x * 2

    @property
    def height(self) -> float:
        return self.label_size - self.padding.y * 2

    @property
    def x1(self) -> float:
        whalf = (self.label_size * self.ratio) / 2
        # whalf = (self.width) / 2
        return self.pos.x + whalf * (self.dirv - 1) + self.padding.x

    @property
    def x2(self) -> float:
        return self.x1 + self.width

    @property
    def y1(self) -> float:
        return self.pos.y - self.height / 2

    @property
    def y2(self) -> float:
        return self.y1 + self.height



# Copyright (c) Kuba Szczodrzyński 2022-05-09.

import json
import re

from pydantic import BaseModel

from .vector import V

if_re = r"<(.+?):(.+?):(.+?):(.+?)>"


class Model(BaseModel):
    class Config:
        arbitrary_types_allowed = True


class EvalFloat(float):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if isinstance(v, str):
            return str_to_num(v)
        return float(v)


def var(s: str, vars: dict) -> str:
    s_prev = s
    while "${" in s:
        for key, value in vars.items():
            s = s.replace(f"${{{key}}}", str(value))
        if s == s_prev:
            # no replacements made anymore
            raise ValueError(f"Missing variables: {s}")
        s_prev = s
    for match in re.findall(if_re, s):
        (var1, var2, true, false) = match
        full = f"<{var1}:{var2}:{true}:{false}>"
        result = true if var1.strip() == var2.strip() else false
        s = s.replace(full, result)
    return s


def str_to_num(s: str) -> float:
    s = s.strip()
    s = eval(s)
    return float(s)


def splitxy(xy: str | tuple | V) -> V:
    if isinstance(xy, tuple):
        return V(xy)
    if isinstance(xy, V):
        return V(xy)
    return V(tuple(map(str_to_num, xy.split(","))))


def merge_dicts(d1, d2):
    if d1 is not None and type(d1) != type(d2):
        raise TypeError(f"d1 and d2 are different types: {d1}, {d2}")
    if isinstance(d2, list):
        if d1 is None:
            d1 = []
        d1.extend(merge_dicts(None, item) for item in d2)
    elif isinstance(d2, dict):
        if d1 is None:
            d1 = {}
        for key in d2:
            d1[key] = merge_dicts(d1.get(key, None), d2[key])
    else:
        d1 = d2
    return d1


def load_json(file: str) -> dict | list:
    with open(file, "r", encoding="utf-8") as f:
        return json.load(f)


# https://stackoverflow.com/a/1094933/9438331
def sizeof(num: int, suffix="iB", base=1024.0) -> str:
    for unit in ["", "K", "M", "G", "T", "P", "E", "Z"]:
        if abs(num) < base:
            return f"{num:.1f} {unit}{suffix}".replace(".0 ", " ")
        num /= base
    return f"{num:.1f} Y{suffix}".replace(".0 ", " ")



# Copyright (c) Kuba Szczodrzyński 2022-06-15.

from enum import Enum


class SectionType(Enum):
    PINS = "Pins"
    LEDS = "LEDs"
    SPI = "SPI Interfaces"
    WIRE = "Wire Interfaces"
    SERIAL = "Serial ports"
    MACROS = "Pin function macros"
    PORTS = "Port availability"
    ARDUINO = "Arduino pin names"



# Copyright (c) Kuba Szczodrzyński 2022-05-12.

from math import radians, tan

from svgwrite import Drawing
from svgwrite.container import Group
from svgwrite.shapes import Rect
from svgwrite.text import Text

from ...utils import EvalFloat
from ...vector import V
from ..base import LabelShape


class Block(LabelShape):
    text: str
    padding: V = V(0.1, 0.2)
    radius: EvalFloat = 0.3
    angle: EvalFloat = 15

    def draw(self, dwg: Drawing):
        g = Group()
        bg = Rect(
            insert=(0, 0),
            size=self.size.tuple,
            rx=self.radius,
            ry=self.radius,
        )
        bg.fill(color=self.color.as_hex())
        bg.skewX(-self.angle)
        skew_len = self.height * tan(radians(self.angle))
        g.add(bg)
        g.translate(self.x1 + skew_len / 2, self.y1)
        dwg.add(g)

        text = self.text
        text_pos = self.center
        text_color = "#423F42" if self.color.as_hsl_tuple()[2] > 0.5 else "white"

        if text.startswith("^"):
            text = text[1:]
            negation_line = Text(
                text="___",
                insert=(text_pos.x, self.y1 - self.width / 16),
                font_family="Consolas",
                font_size=1.2,
                text_anchor="middle",
                dominant_baseline="middle",
            )
            negation_line.fill(color=text_color)
            dwg.add(negation_line)
            text_pos.y += self.width / 32

        txt = Text(
            text=text,
            insert=text_pos.tuple,
            font_family="Consolas",
            font_size=1.2,
            text_anchor="middle",
            dominant_baseline="central",
        )
        txt.fill(color=text_color)
        dwg.add(txt)



# Copyright (c) Kuba Szczodrzyński 2022-05-12.

from svgwrite import Drawing
from svgwrite.shapes import Rect

from ...models.enums import IOType
from ...vector import V
from ..base import LabelShape


class IOLine(LabelShape):
    type: IOType
    padding: V = V(0.5, 0.5)

    def draw(self, dwg: Drawing):
        rect = Rect(
            insert=self.pos1.tuple,
            size=self.size.tuple,
        )
        rect.fill(color=self.color.as_hex())
        dwg.add(rect)

    @property
    def height(self) -> float:
        return 0.2



# Copyright (c) Kuba Szczodrzyński 2022-05-11.

from .board import Board
from .enums import IOType, LabelDir, RoleType, RoleValue, ShapeType, Side
from .flash_region import FlashRegion
from .pcb import Pcb
from .role import Role
from .template import Template

__all__ = [
    "ShapeType",
    "Side",
    "LabelDir",
    "RoleType",
    "Pcb",
    "Role",
    "Template",
    "FlashRegion",
    "Board",
    "IOType",
    "RoleValue",
]



# Copyright (c) Kuba Szczodrzyński 2022-05-12.

from abc import ABC
from glob import glob
from os.path import basename, isfile, join
from typing import Callable, Optional

from ..utils import load_json, merge_dicts


class CoreCache(ABC):
    _cache: dict[str, dict[str, dict]] = {
        "boards": {},
        "board_objs": {},
        "shapes": {},
        "templates": {},
    }
    json_hook: Optional[Callable[[str, str, dict, Optional[str]], None]] = None

    def get_dirs(self, type: str) -> list[str]:
        attr_name = f"_dirs_{type}"
        if not hasattr(self, attr_name):
            return []
        return getattr(self, attr_name)

    def list_json(self, type: str) -> set[str]:
        dirs = self.get_dirs(type)
        files = set()
        for dir in dirs:
            for file in glob(join(dir, "*.json")):
                files.add(basename(file).rpartition(".")[0])
        return files

    def load_json(self, type: str, name: str) -> dict | None:
        if name in self._cache[type]:
            if self.json_hook:
                self.json_hook(type, name, self._cache[type][name], None)
            return self._cache[type][name]
        dirs = self.get_dirs(type)
        for dir in dirs:
            file = join(dir, f"{name}.json")
            if isfile(file):
                data = load_json(file)
                self._cache[type][name] = data
                if self.json_hook:
                    self.json_hook(type, name, data, file)
                return data
        return None

    def load_shape(self, name: str) -> dict:
        return self.load_json("shapes", name)

    def load_board_base(self, name: str) -> dict:
        name = join("_base", name)
        return self.load_json("boards", name)

    def load_board(self, name: str, allow_cache: bool = True) -> dict:
        if allow_cache and name in self._cache["board_objs"]:
            return self._cache["board_objs"][name]
        manifest = self.load_json("boards", name)
        if "_base" in manifest:
            bases = manifest["_base"]
            if not isinstance(bases, list):
                bases = [bases]

            result = {}
            for base in bases:
                base_manifest = self.load_board_base(base)
                merge_dicts(result, base_manifest)
            merge_dicts(result, manifest)
            manifest = result
        self._cache["board_objs"][name] = manifest
        return manifest

    def load_template(self, name: str) -> dict:
        return self.load_json("templates", name)



# Copyright (c) Kuba Szczodrzyński 2022-05-11.

from ..utils import Model, sizeof
from .flash_region import FlashRegion
from .pcb import Pcb


class Board(Model):
    build: "BoardBuild"
    connectivity: list[str] = None
    debug: "BoardDebug" = None
    doc: "BoardDoc" = None
    flash: list[FlashRegion] = None
    name: str
    pcb: Pcb = None
    upload: "BoardUpload"
    url: str = None
    vendor: str

    @property
    def id(self) -> str:
        return self.build.variant

    @property
    def size_flash(self) -> str:
        return sizeof(self.upload.flash_size)

    @property
    def size_firmware(self) -> str:
        return sizeof(self.upload.maximum_size)

    @property
    def size_ram(self) -> str:
        return sizeof(self.upload.maximum_ram_size)

    @property
    def cpu_freq(self) -> str:
        f_cpu = "".join(c for c in self.build.f_cpu if c.isnumeric())
        return sizeof(int(f_cpu), suffix="Hz", base=1000)

    @property
    def has_arduino_core(self) -> bool:
        return True


class BoardBuild(Model):
    f_cpu: str
    family: str
    mcu: str
    variant: str


class BoardDebug(Model):
    protocol: str
    protocols: list[str]


class BoardDoc(Model):
    params: "BoardDocParams" = None
    links: dict[str, str] = None
    extra: list[str] = None
    fccid: str = None


class BoardDocParams(Model):
    extra: dict[str, str] = None
    manufacturer: str = None
    series: str = None
    voltage: str = None


class BoardUpload(Model):
    flash_size: int
    maximum_ram_size: int
    maximum_size: int
    protocol: str = None
    protocols: list[str] = None
    speed: int = None


Board.update_forward_refs()
BoardDoc.update_forward_refs()



# Copyright (c) Kuba Szczodrzyński 2022-05-12.

from pydantic.color import Color
from svgwrite import Drawing
from svgwrite.gradients import LinearGradient
from svgwrite.mixins import Presentation

from ..utils import EvalFloat, Model
from ..vector import V


class FillStyle(Model):
    color: Color = None
    lgrad: tuple[V, Color, V, Color] = None
    width: EvalFloat = None

    def apply_to(self, dwg: Drawing, el: Presentation, stroke: bool = False):
        color = None
        if self.color:
            color = self.color.as_hex()
        elif self.lgrad:
            unit = max(*self.lgrad[0].tuple, *self.lgrad[2].tuple)
            grad = LinearGradient(
                start=self.lgrad[0].tuple,
                end=self.lgrad[2].tuple,
                gradientUnits="userSpaceOnUse" if unit > 1 else "objectBoundingBox",
            )
            grad.add_stop_color(offset="0%", color=self.lgrad[1])
            grad.add_stop_color(offset="100%", color=self.lgrad[3])
            color = grad
            dwg.add(grad)
        if color:
            if stroke:
                if not self.width:
                    raise ValueError("No stroke width")
                el.stroke(color=color, width=self.width)
            else:
                el.fill(color=color)



# Copyright (c) Kuba Szczodrzyński 2022-06-15.

from abc import ABC

from natsort import natsort_keygen

from ..models import Role
from .features import PinFeatures
from .section import SectionType

SectionItem = tuple[int, str | None]


class VariantParts(ABC):
    # { name: (c_name, features, comment, roles) }
    pins: dict[str, tuple[str, PinFeatures, str]]
    # { pin_number: arduino_name }
    gpio_map: dict[int, str]
    # { section_type: { key: (value, comment) } }
    sections: dict[SectionType, dict[str, SectionItem]]
    # { name: pin_name }
    static_pins: dict[str, str]

    def add_pin(self, name: str, gpio: str, comment: str = None) -> bool:
        if name in self.pins:
            return False
        self.pins[name] = (gpio, PinFeatures.PIN_NONE, comment)
        return True

    def add_pin_feature(self, name: str, feature: PinFeatures):
        if name not in self.pins:
            return
        (gpio, features, comment) = self.pins[name]
        features &= ~PinFeatures.PIN_NONE
        features |= feature
        self.pins[name] = (gpio, features, comment)

    def add_item(
        self,
        section: SectionType,
        key: str,
        value: int,
        comment: str = None,
    ):
        if section not in self.sections:
            self.sections[section] = {}
        self.sections[section][key] = (value, comment)

    def increment_item(self, section: SectionType, key: str):
        if section not in self.sections:
            self.sections[section] = {}
        if key in self.sections[section]:
            value, comment = self.sections[section][key]
            self.sections[section][key] = (value + 1, comment)
        else:
            self.sections[section][key] = (1, None)

    def format_pins(self) -> str:
        name_list = "lt_arduino_pin_info_list"
        name_map = "lt_arduino_pin_gpio_map"
        out = [
            "// clang-format off",
            f"PinInfo {name_list}[PINS_COUNT] = {{",
        ]

        def sort_pins(item):
            # (name, (c_name, features, comment))
            return int(item[0][1:]) + (1000 if item[0][0] == "D" else 2000)

        pins: list[tuple[str, tuple[str, PinFeatures, str]]]
        # noinspection PyTypeChecker
        pins = sorted(self.pins.items(), key=sort_pins)
        gpio_map = sorted(self.gpio_map.items())

        pin_name_map: dict[str, tuple[str, int]] = {}
        for index, (arduino_name, (c_name, _, _)) in enumerate(pins):
            pin_name_map[arduino_name] = c_name, index

        len_c_name = 0
        len_features = 0
        for _, (c_name, features, _) in pins:
            features = str(features).partition(".")[2].replace("|", " | ")
            len_c_name = max(len_c_name, len(c_name))
            len_features = max(len_features, len(features))

        # (name, (c_name, features, comment))
        for name, (c_name, features, comment) in pins:
            features = str(features).partition(".")[2]
            features = " | ".join(reversed(features.split("|")))
            out.append(f"\t// {name}: {comment}")
            pad_c_name = " " * (len_c_name - len(c_name))
            pad_features = " " * (len_features - len(features))
            out.append(
                f"\t{{{c_name}, {pad_c_name}{features}, {pad_features}PIN_NONE, 0}},"
            )
        out.append("};\n")
        out.append(f"PinInfo *{name_map}[] = {{")

        max_pin = max(pin for pin, _ in gpio_map)
        max_idx = max(idx for _, (_, idx) in pin_name_map.items())
        len_pin = len(str(max_pin))
        len_idx = len(str(max_idx))

        for pin_number, arduino_name in gpio_map:
            c_name, index = pin_name_map[arduino_name]
            pad_pin = " " * (len_pin - len(str(pin_number)))
            pad_idx = " " * (len_idx - len(str(index)))
            out.append(
                f"\t[{pin_number}]{pad_pin} = "
                f"&({name_list}[{index}]), {pad_idx}"
                f"// {c_name} ({arduino_name})",
            )

        out.append("};")
        out.append("// clang-format on")
        return "\n".join(out)

    def format_sections(self) -> str:
        out = [
            "// clang-format off",
            "",
        ]

        def sort_pins(item):
            # (key, (value, comment))
            pin = item[0]
            shift = 10000
            if pin.startswith("PIN_"):
                pin = pin[4:]
                shift = 0
            return int(pin[1:]) + (1000 if pin[0] == "D" else 2000) + shift

        natsort_key = natsort_keygen()

        for section_type in SectionType:
            if section_type not in self.sections:
                continue
            section = self.sections[section_type].items()
            out.append(f"// {section_type.value}")
            out.append(f"// {len(section_type.value) * '-'}")
            max_key = 0
            max_value = 0

            for key, (value, _) in section:
                value = str(value)
                if key.startswith("PIN_"):
                    value = f"{value}u"
                max_key = max(max_key, len(key))
                max_value = max(max_value, len(value))

            if section_type == SectionType.ARDUINO:
                section = sorted(section, key=sort_pins)
            elif section_type != SectionType.PINS:
                section = sorted(section, key=natsort_key)

            for key, (value, comment) in section:
                value = str(value)
                if key.startswith("PIN_"):
                    value = f"{value}u"
                if comment:
                    comment = f"// {comment}"
                else:
                    comment = ""
                pad_key = " " * (max_key - len(key))
                pad_value = " " * (max_value - len(value))
                out.append(
                    f"#define {key}{pad_key} {value}{pad_value} {comment}".strip()
                )
            out.append("")

        heading = "Static pin names"
        out.append(f"// {heading}")
        out.append(f"// {len(heading) * '-'}")
        for name, pin_name in sorted(self.static_pins.items(), key=natsort_key):
            out.append(f"static const unsigned char {name} = {pin_name};")
        out.append("")

        return "\n".join(out)



# Copyright (c) Kuba Szczodrzyński 2022-05-11.

import json
import re
from copy import deepcopy
from importlib.metadata import version
from os.path import dirname, isfile, join

from ..mixins import HasId, ParentType
from ..models import Board, FlashRegion, Pcb, RoleType, ShapeType, Side, Template
from ..models.enums import RoleValue
from ..shapes.base import Shape
from ..shapes.circle import Circle
from ..shapes.group import ShapeGroup
from ..shapes.label import Label
from ..shapes.rect import Rect
from ..shapes.text import Text
from ..utils import var
from ..vector import V
from .cache import CoreCache
from .getters import CoreGetters


class Core(CoreCache, CoreGetters):
    shape_ctors: dict[ShapeType, type]
    is_libretiny: bool = False

    _dir_base: str
    _dirs_boards: list[str]
    _dirs_shapes: list[str]
    _dirs_templates: list[str]

    def __init__(self) -> None:
        self._dir_base = join(dirname(__file__), "..", "res")
        self._dirs_boards = [
            join(dirname(__file__), "..", "..", "..", "..", "boards"),
            join(self._dir_base, "boards"),
            "boards",
        ]
        self._dirs_shapes = [
            join(self._dir_base, "shapes"),
        ]
        self._dirs_templates = [
            join(self._dir_base, "templates"),
        ]
        self._file_presets = join(self._dir_base, "presets.json")
        self._file_roles = join(self._dir_base, "roles.json")
        self._file_flash = join(self._dir_base, "flash.json")
        self.shape_ctors = {
            ShapeType.RECT: Rect,
            ShapeType.CIRCLE: Circle,
            ShapeType.SUBSHAPE: ShapeGroup,
            ShapeType.TEXT: Text,
        }
        self.is_libretiny = isfile(
            join(dirname(__file__), "..", "..", "..", "..", "platform.json")
        ) or isfile("families.json")

    @property
    def version(self) -> str | None:
        pyproject = join(dirname(__file__), "..", "..", "pyproject.toml")
        if isfile(pyproject):
            with open(pyproject, "r", encoding="utf-8") as f:
                text = f.read()
                ver = re.search(r"version\s?=\s?\"(.+?)\"", text)
                if ver:
                    return ver.group(1)
        try:
            return version("boardgen")
        except Exception:
            return None

    def add_custom_dirs(
        self,
        boards: list[str] | str = None,
        shapes: list[str] | str = None,
        templates: list[str] | str = None,
    ):
        if isinstance(boards, str):
            boards = [boards]
        if isinstance(shapes, str):
            shapes = [shapes]
        if isinstance(templates, str):
            templates = [templates]

        if boards:
            self._dirs_boards = boards + self._dirs_boards
        if shapes:
            self._dirs_shapes = shapes + self._dirs_shapes
        if templates:
            self._dirs_templates = templates + self._dirs_templates

    def add_custom_json(
        self,
        presets: dict = None,
        roles: dict = None,
        flash: dict = None,
    ):
        if presets:
            self.presets.update(presets)
        if roles:
            self.roles.update(roles)
        if flash:
            self.flash.update(flash)

    def build_shapes(self, name: str, parent: ParentType, pos: V = None) -> list[Shape]:
        """Load the specified shape JSON into a list of Shape objects.

        Args:
            name (str): Shape name.
            parent (ParentType): Parent object (Shape, Pcb, etc).
            pos (V | None): Move the shape by the vector. Defaults to None.
        """
        shape = self.load_shape(name)
        return [self.build_shape(parent, data, pos) for data in shape]

    def build_shape(self, parent: ParentType, data: dict, pos: V = None) -> Shape:
        """Deserialize a single shape from JSON.

        Args:
            parent (ParentType): Parent object.
            data (dict): Input JSON data.
            pos (V, optional): Move the shape by the vector. Defaults to None.
        """
        shape = Shape.deserialize(self, parent, data)
        if pos:
            shape.move(pos)
        return shape

    def build_presets(self, vars: dict) -> dict[str, dict]:
        """Return all defined presets with specified vars applied.

        Args:
            vars (dict): Variables to apply. May be None.
        """
        if not vars:
            return self.presets
        # ugly way to replace all vars
        presets = json.dumps(self.presets)
        presets = var(presets, vars)
        return json.loads(presets)

    def get_board(self, name: str) -> Board:
        """Load and build the specified board.

        Args:
            name (str): Board name.
        """
        manifest = self.load_board(name)
        manifest = deepcopy(manifest)
        pcb = manifest.get("pcb", None)
        pinout = pcb.get("pinout", None) if pcb else None
        ic_pins = pcb.get("ic", None) if pcb else None

        if pinout:
            for roles in pinout.values():
                roles: dict[str, RoleValue]
                # add IC pinout mapping from board manifest
                ic_pin = str(roles.get("IC", None))
                if ic_pins and ic_pin in ic_pins:
                    roles |= ic_pins[ic_pin]

                for role_type, functions in dict(roles).items():
                    if role_type == "ARD":
                        digital = []
                        analog = []
                        if not isinstance(functions, list):
                            functions = [str(functions)]
                        for function in functions:
                            match function[0]:
                                case "D":
                                    digital.append(function)
                                case "A":
                                    analog.append(function)
                        if digital:
                            if len(digital) != 1:
                                raise ValueError(f"Invalid digital pins: {digital}")
                            roles["ARD_D"] = digital[0]
                        if analog:
                            if len(analog) != 1:
                                raise ValueError(f"Invalid analog pins: {analog}")
                            roles["ARD_A"] = analog[0]
                        roles.pop("ARD", None)

        if "flash" in manifest and isinstance(manifest["flash"], dict):
            flash: list[FlashRegion] = []
            hex_offs_len = 0
            hex_size_len = 0
            for region, layout in manifest["flash"].items():
                (offset, _, length) = layout.partition("+")
                offset = int(offset, 16)
                length = int(length, 16)
                hex_offs_len = max(hex_offs_len, len(hex(offset)))
                hex_size_len = max(hex_size_len, len(hex(length)))
                desc = self.flash.get(region, "(unknown)")
                if not isinstance(desc, dict):
                    desc = dict(title=desc)
                flash.append(
                    FlashRegion(
                        name=region,
                        offset=offset,
                        length=length,
                        **desc,
                    )
                )
            for region in flash:
                region.hex_offs_len = hex_offs_len - 2
                region.hex_size_len = hex_size_len - 2
            manifest["flash"] = flash

        board = Board(**manifest)
        pcb = board.pcb

        if not pcb:
            return board
        pcb.vars["NAME"] = board.build.variant
        pcb.vars["TITLE"] = board.name
        pcb.vars["SYMBOL"] = pcb.symbol

        shapes = {
            Side.FRONT: [],
            Side.BACK: [],
        }
        sources: list[HasId] = []
        for template_name in pcb.templates:
            template = Template(**self.load_template(template_name))
            template.vars |= pcb.vars
            pcb.pads |= template.pads
            pcb.test_pads |= template.test_pads
            sources.append(template)
        sources.append(pcb)

        for side in Side:
            for parent in sources:
                parent.id_suffix = side.value
                shapes[side] += [
                    self.build_shape(parent, data)
                    for data in getattr(parent, side.value)
                ]
                parent.id_suffix = None

        for side, items in shapes.items():
            pcb.shapes[side] = ShapeGroup.wrap(self, name + "." + side.value, items)

        return board

    def build_labels(self, pcb: Pcb, side: Side) -> tuple[list[Label], V, V]:
        pads = pcb.pads
        pads |= pcb.test_pads
        pins = pcb.pinout
        hidden = pcb.pinout_hidden.split(",")

        x1, y1, x2, y2 = (None, None, None, None)
        labels: list[Label] = []
        for pin, roles in pins.items():
            if pin not in pads:
                continue
            pad = pcb.shapes[side].get_by_id(pads[pin])
            if not pad:
                continue
            label = Label(
                pos=pad.pos,
                role_type=RoleType.NC,
                ratio=1.0,
                color="#000",
                roles=roles,
            )
            label.build(self, pin, pad, hidden)
            if x1 is None:
                x1 = label.pos1.x
                y1 = label.pos1.y
                x2 = label.pos2.x
                y2 = label.pos2.y
            else:
                x1 = min(x1, label.pos1.x)
                y1 = min(y1, label.pos1.y)
                x2 = max(x2, label.pos2.x)
                y2 = max(y2, label.pos2.y)
            labels.append(label)

        return (labels, V(x1, y1), V(x2, y2))



#  Copyright (c) Kuba Szczodrzyński 2023-6-3.

import json
from dataclasses import dataclass
from logging import debug
from os.path import join

import wx
import wx.adv
import wx.xrc
from ltchiptool.gui.panels.base import BasePanel
from ltchiptool.gui.utils import with_event
from ltchiptool.util.lvm import LVM

from boardgen import Core, HasVars, V
from boardgen.draw_util import draw_shapes, get_pcb_images
from boardgen.models import Board, Template
from boardgen.shapes import Shape, ShapeGroup

from .svg import SvgPanel
from .utils import jsonpath


@dataclass
class EditableShape:
    prefix: str
    data: dict
    shape: Shape = None

    @staticmethod
    def build(core: Core, _vars: dict, data: dict) -> "EditableShape":
        _vars |= {"DUMMY": 0}
        obj = EditableShape(
            prefix="",
            data=data,
        )
        obj.rebuild(core, _vars)
        return obj

    def rebuild(self, core: Core, _vars: dict) -> None:
        self.shape = core.build_shape(parent=HasVars(vars=_vars), data=self.data)

    @property
    def title(self) -> str:
        title = self.prefix + type(self.shape).__name__ + str(self.shape.pos.tuple)
        if self.shape.id:
            title += " - " + self.shape.id
        if isinstance(self.shape, ShapeGroup):
            title = "Shape - " + self.shape.name
        return title


class BoardgenPanel(BasePanel):
    file_map: dict[str, str] = None
    draw_items: dict[str, tuple[str, str]] = None
    draw_object: Board | list[Shape] | Shape
    edit_items: dict[str, tuple[str, dict]] = None
    edit_selection: str = None
    _vars: dict

    def __init__(self, parent: wx.Window, frame):
        super().__init__(parent, frame)
        self.LoadXRCFile("boardgen.xrc")

        self._shapes = []
        self._vars = {}

        self.Sizer = wx.BoxSizer(wx.VERTICAL)
        self.Splitter = wx.SplitterWindow(self, style=wx.SP_3D | wx.SP_LIVE_UPDATE)
        self.Left: wx.Panel = self.Xrc.LoadPanel(self.Splitter, "PreviewPanel")
        self.Right: wx.Panel = self.Xrc.LoadPanel(self.Splitter, "EditPanel")
        self.Splitter.SetMinimumPaneSize(300)
        self.Splitter.SetSashGravity(0.3)
        self.Splitter.SplitVertically(self.Left, self.Right, sashPosition=-600)
        self.Sizer.Add(self.Splitter, proportion=1, flag=wx.EXPAND)
        self.SetSizer(self.Sizer)

        self.AddToNotebook("boardgen")

        self.PreviewPanel: wx.Panel = self.Left.FindWindowByName("preview_panel")
        self.PreviewBox: wx.BoxSizer = self.PreviewPanel.GetSizer()
        self.Svg = SvgPanel(self.PreviewPanel)
        self.PreviewBox.Add(self.Svg, proportion=1, flag=wx.EXPAND)

        self.EditItem = self.BindComboBox("combo_edit_item")
        self.DrawItem = self.BindComboBox("combo_preview_item")
        self.Data: wx.TextCtrl = self.FindWindowByName("input_data")
        self.Data.Bind(wx.EVT_TEXT, self.OnDataText)
        self.Data.Bind(wx.EVT_KEY_UP, self.OnDataPosition)
        self.Data.Bind(wx.EVT_LEFT_UP, self.OnDataPosition)
        self.Vars = self.BindTextCtrl("input_vars")
        self.Vars.Bind(wx.EVT_TEXT, self.OnVarsText)
        self.Path = self.BindTextCtrl("input_path")
        self.Error = self.BindTextCtrl("input_error")

        self.lvm = LVM.get()
        self.core = Core()
        self.core.add_custom_dirs(boards=join(self.lvm.path(), "boards"))
        self.core.json_hook = self.AddEditItem

        self.file_map = {}
        self.draw_items = {}
        self.edit_items = {}

        self.ReloadLists()

    def GetSettings(self) -> dict:
        return dict()

    def SetSettings(
        self,
        **_,
    ) -> None:
        pass

    def ReloadLists(self) -> None:
        draw_selection = self.DrawItem.GetStringSelection()
        items = {}

        for board_name in self.core.list_json("boards"):
            board = self.core.load_board(board_name)
            items[f"Board - {board['name']}"] = "board", board_name
        for template_name in self.core.list_json("templates"):
            template = self.core.load_template(template_name)
            items[f"Template - {template['title']}"] = "template", template_name
        for shape_name in self.core.list_json("shapes"):
            # shape = self.core.load_shape(shape_name)
            items[f"Shape - {shape_name}"] = "shape", shape_name

        self.DrawItem.SetItems(sorted(items.keys()))
        self.DrawItem.SetStringSelection(draw_selection)
        self.draw_items = items

    def OnUpdate(self, target: wx.Window = None) -> None:
        super().OnUpdate(target)
        if not self.IsShown():
            return
        match target:
            case self.DrawItem:
                self.Svg.ClearSvg()
                self.UpdateDrawItem(force_list=True)
            case self.EditItem:
                self.UpdateEditItem()

            case _:
                # debug(f"OnUpdate({target})")
                pass

    def UpdateDrawItem(self, force_list: bool = False) -> None:
        item_type, item_name = self.draw_item

        self.edit_items = {}
        obj: dict = {}
        try:
            match item_type:
                case "board":
                    obj = self.core.load_board(item_name, allow_cache=False)
                case "template":
                    obj = self.core.load_template(item_name)
                case "shape":
                    obj = self.core.load_shape(item_name)
        except Exception as e:
            self.SetError(e)
            return
        self.vars = obj["vars"] if "vars" in obj else {}
        parent = HasVars(vars=self.vars)

        try:
            match item_type:
                case "board":
                    self.draw_object = self.core.get_board(item_name)

                case "template":
                    front: list[Shape] = []
                    back: list[Shape] = []
                    template = Template(**obj)
                    for data in template.front:
                        shape = self.core.build_shape(parent=parent, data=data)
                        front.append(shape)
                    for data in template.back:
                        shape = self.core.build_shape(parent=parent, data=data)
                        back.append(shape)
                    if front and back:
                        self.draw_object = [
                            ShapeGroup.wrap(self.core, item_name, front),
                            ShapeGroup.wrap(self.core, item_name, back),
                        ]
                    elif front:
                        self.draw_object = ShapeGroup.wrap(self.core, item_name, front)
                    elif back:
                        self.draw_object = ShapeGroup.wrap(self.core, item_name, back)

                case "shape":
                    shapes: list[Shape] = []
                    for data in obj:
                        shape = self.core.build_shape(parent=parent, data=data)
                        shapes.append(shape)
                    self.draw_object = ShapeGroup.wrap(self.core, item_name, shapes)

            self.Redraw()
            self.SetError(None)
        except Exception as e:
            self.SetError(e)
            if not force_list:
                return

        selection = self.EditItem.GetStringSelection()
        self.EditItem.Clear()
        self.EditItem.Append(list(self.edit_items.keys()))
        self.EditItem.SetStringSelection(selection)

        if not self.EditItem.GetStringSelection():
            if self.EditItem.GetItems():
                self.edit_item = self.EditItem.GetItems()[0]
            else:
                self.Data.Clear()
                self.Path.Clear()

    def AddEditItem(self, type: str, name: str, data: dict, file: str | None) -> None:
        key = f"{type} -> {name}".replace("\\", "/")
        value = f"{type}/{name}", data
        self.edit_items[key] = value
        if file:
            self.file_map[value[0]] = file

    def UpdateEditItem(self) -> None:
        item_name, obj = self.edit_item
        self.edit_selection = item_name
        self.Data.ChangeValue(json.dumps(obj, indent=4))
        if "pcb" in obj:
            obj = obj["pcb"]
        self.vars = obj["vars"] if "vars" in obj else {}

    def Redraw(self) -> None:
        debug(f"Draw object: {self.draw_object}")
        match self.draw_object:
            case Board():
                images = get_pcb_images(
                    self.core,
                    self.draw_object.pcb,
                    with_labels=True,
                )
            case list():
                images = self.draw_object
            case Shape():
                images = [self.draw_object]
            case _:
                return
        if not images:
            self.Svg.ClearSvg()
            return
        dwg = draw_shapes(V(1024, 500), 12, images, with_canvas=False)
        self.Svg.LoadSvg(dwg.tostring())

    @with_event
    def OnDataText(self, event: wx.Event) -> None:
        event.Skip()
        try:
            text = self.Data.GetValue()
            obj = json.loads(text)
            if obj:
                self.data = obj
            self.UpdateDrawItem()
        except json.JSONDecodeError:
            return

    @with_event
    def OnVarsText(self, event: wx.Event) -> None:
        event.Skip()
        self.UpdateDrawItem()

    @with_event
    def OnDataPosition(self, event: wx.Event) -> None:
        event.Skip()
        text = self.Data.GetValue()
        text = text.replace("\n", "\r\n")
        pos = max(0, self.Data.GetInsertionPoint() - 1)
        path = jsonpath(text, pos)
        if path:
            self.Path.ChangeValue(path.replace(".", " -> "))
        else:
            self.Path.ChangeValue("")

    def SetError(self, e: Exception | None) -> None:
        if not e:
            self.Error.Clear()
        else:
            self.Error.SetValue(f"{type(e).__name__}: {e}")

    @property
    def vars(self) -> dict[str, str]:
        lines: list[str] = self.Vars.GetValue().splitlines()
        data = {}
        for line in lines:
            key, _, value = line.strip().partition("=")
            data[key] = value
        data |= self._vars
        return data

    @vars.setter
    def vars(self, value: dict[str, str]) -> None:
        self._vars = value

    @property
    def draw_item(self) -> tuple[str, str] | None:
        return self.draw_items.get(self.DrawItem.GetStringSelection(), None)

    @draw_item.setter
    def draw_item(self, value: str) -> None:
        for key, item in self.draw_items.items():
            if key == value or item[1] == value:
                self.DrawItem.SetStringSelection(key)
                self.UpdateDrawItem()
                return

    @property
    def edit_item(self) -> tuple[str, dict] | None:
        return self.edit_items.get(self.EditItem.GetStringSelection(), None)

    @edit_item.setter
    def edit_item(self, value: str) -> None:
        for key, item in self.edit_items.items():
            if key == value or item[1] == value == value:
                self.EditItem.SetStringSelection(key)
                self.UpdateEditItem()
                return

    @property
    def data(self) -> dict | None:
        item = self.edit_item
        return item and item[1]

    @data.setter
    def data(self, value: dict) -> None:
        data = self.data
        if data is None:
            return
        debug(f"Data change, id={id(data)}")
        data.clear()
        if isinstance(data, dict):
            data.update(value)
        if isinstance(data, list):
            data += value
